
class RoboDocInfo:
	"""
	Class which reads text templates from a file.

	The file to be read must be in the format:

	@method_name(param1,param2,....)
	text which may contain format instructions like %(param1)s etc.

	The generated object will have methods with the same names and parameters
	iterating over the formatted lines

	EXAMPLE

	-------- file test.txt -------------
	@foo(name, age)
	Mr. %(name)s is %(age)d years old.
	@baz(name,size,cdate,mdate)
	The file %(name)s has a size of %(size)5.2f MB.
	It has been created on %(cdate)s.
	Its last modification was on %(mdate).
	---------------------------------------

	>>> DOC = RoboDoc('test.txt')
	>>> for line in DOC.baz("autoexec.bat", 3.053, "1980-01-01", "1981-03-17"):
	... 	print line
	The file autoexec.bat has a size of  3.05 MB.
	It has been created on 1980-01-01.
	Its last modification was on 1981-03-17.
	"""
	def __init__(self, fname):
		"""
		Create a new object from a file

		PARAMETER

		   fname -- the name of the file to be read.
		"""
		self._info = {}
		self._args = {}
		self._read(fname)

	def _read(self, fname):
		f = open(fname, "r")
		key = None
		for line in f.readlines():
			if line.startswith("@"):
				if "(" in line:
					tmp = line.split("(")
					line = tmp[0]
					args = map(lambda s: s.strip(),
							tmp[1].strip(" )\t\r\n").split(","))
				else:
					args = []
				key = line[1:].strip()
				self._args[key] = args
				self._info[key] = []
			elif key is not None:
				self._info[key].append(line.rstrip())
		f.close()
	
	def __getattr__(self, name):
		if name in self._info:
			lines = self._info[name]
			param = self._args[name]

			def meth(*args):
				if len(param) != len(args):
					raise TypeError, \
						"RoboDocInfo.%s(%s) expects exactly %d arguments (%d given)" \
							% (name, ", ".join(list(param)), len(param), len(args))

				opts = dict(zip(param, args))
				for line in lines:
					yield line % opts

			return meth
		else:
			raise AttributeError, \
				"This RoboDocInfo does not contain an entry for %r." % name
